/**********************************************************************
	Copyright (c) 1996 Immersion Corporation

	Permission to use, copy, modify, distribute, and sell this
	software and its documentation may be granted without fee;
	interested parties are encouraged to request permission from
		Immersion Corporation
		2158 Paragon Drive
		San Jose, CA 95131
		408-467-1900

	IMMERSION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
	IN NO EVENT SHALL IMMERSION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
	LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
	CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

	FILE: READCFG.C
	CONTENTS: This is the code that will read the configuration
		file and get the data from the file.
	STARTED: 6/28/96
	NOTES/REVISIONS:
	COMPILER DEFINITIONS:
**********************************************************************/

	/* === Compiler Definitions === */

	/* === Includes === */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "readcfg.h"

	/* === Constants and Definitions === */
#define RC_TYPE_INT				1
#define RC_TYPE_LONG				2
#define RC_TYPE_FLOAT			3
#define RC_TYPE_DOUBLE			4	
#define RC_TYPE_HEX				5

	/* === Data Structures === */
typedef union _unDataUnion {
	int i;
	long l;
	float f;
	double d;
	int h;
} unDataUnion;

typedef struct _stCfgEntry {
	char *name;
	int type;
	void *var;
	unDataUnion deflt;
	struct _stCfgEntry *next;
} stCfgEntry;

	/* === External Access Globals === */

	/* === Local Access Globals === */
static stCfgEntry *cfgFirst = NULL;
static stCfgEntry *cfgLast = NULL;

	/* === Local Function Prototypes === */
stCfgEntry *CreateCfgEntry(char *name, int type, void *var);
void AddCfgEntry(stCfgEntry *ne);


/**********************************************************************
	FUNCTION: ReadCfgAdd[Int/Long/Float/Double/Hex]
	CONTENTS: This will add an entry to the list to read from the data
		file
	PARAMETERS:
		name -- The text name to read from the file.
		var -- The location to store the data.
		deflt -- The default value for the data.
	RETURNS:
		int -- Flag for success (0 == failure/1 == success)
	NOTES:
**********************************************************************/
int ReadCfgAddInt(char *name, int *var, int deflt)
{
	stCfgEntry *ne;

		/* allocate the new entry */
	if ((ne = CreateCfgEntry(name,RC_TYPE_INT,var)) == NULL)
		return 0;

	ne->deflt.i = deflt;

	return 1;
}

int ReadCfgAddLong(char *name, long *var, long deflt)
{
	stCfgEntry *ne;

		/* allocate the new entry */
	if ((ne = CreateCfgEntry(name,RC_TYPE_LONG,var)) == NULL)
		return 0;

	ne->deflt.l = deflt;

	return 1;
}

int ReadCfgAddFloat(char *name, float *var, float deflt)
{
	stCfgEntry *ne;

		/* allocate the new entry */
	if ((ne = CreateCfgEntry(name,RC_TYPE_FLOAT,var)) == NULL)
		return 0;

	ne->deflt.f = deflt;

	return 1;
}

int ReadCfgAddDouble(char *name, double *var, double deflt)
{
	stCfgEntry *ne;

		/* allocate the new entry */
	if ((ne = CreateCfgEntry(name,RC_TYPE_DOUBLE,var)) == NULL)
		return 0;

	ne->deflt.d = deflt;

	return 1;
}

int ReadCfgAddHex(char *name, int *var, int deflt)
{
	stCfgEntry *ne;

		/* allocate the new entry */
	if ((ne = CreateCfgEntry(name,RC_TYPE_HEX,var)) == NULL)
		return 0;

	ne->deflt.h = deflt;

	return 1;
}


/**********************************************************************
	FUNCTION: ReadCfgExecute
	CONTENTS: This will actually read the given configuration file.
	PARAMETERS:
		filename -- The file to read.
		flags	-- The flags for the read.
	RETURNS:
		int -- The number of items read.  A value less than 0 indicates
			an error.
	NOTES:
		If 2 items have the same text name, they will be assigned the
			same values.
**********************************************************************/
int ReadCfgExecute(char *filename,unsigned int flags)
{
	FILE *fin;
	char buffer[200];		/* the is oversized on purpose */
	char *nindex, *vindex;
	stCfgEntry *cur;
	int read = 0;

		/* assign the defaults */
	if (flags & RCE_APPLY_DEFAULTS) {
		cur = cfgFirst;
		while (cur != NULL) {
			switch (cur->type) {
				case RC_TYPE_INT : *((int *) (cur->var)) = cur->deflt.i; break;
				case RC_TYPE_LONG	: *((long *) (cur->var)) = cur->deflt.l; break;
				case RC_TYPE_FLOAT : *((float *) (cur->var)) = cur->deflt.f; break;
				case RC_TYPE_DOUBLE : *((double *) (cur->var)) = cur->deflt.d; break;
				case RC_TYPE_HEX : *((int *) (cur->var)) = cur->deflt.h; break;
			}
			cur = cur->next;
		}
	}

		/* open the file */
	if ((fin = fopen(filename,"r")) == NULL)
		return -1;

		/* read the new values */
	while (fgets(buffer,sizeof(buffer),fin) != NULL && strlen(buffer) > 0) {
		nindex = strtok(buffer," =");
		vindex = strtok(NULL," =");

		cur = cfgFirst;
		while (cur != NULL) {
			if (strcmp(cur->name,nindex) == 0) {
				switch (cur->type) {
					case RC_TYPE_INT : *((int *) (cur->var)) = atoi(vindex); break;
					case RC_TYPE_LONG	: *((long *) (cur->var)) = atol(vindex); break;
					case RC_TYPE_FLOAT : *((float *) (cur->var)) = atof(vindex); break;
					case RC_TYPE_DOUBLE : *((double *) (cur->var)) = atof(vindex); break;
					case RC_TYPE_HEX : *((int *) (cur->var)) =
               	(int) strtol(vindex,(char**)NULL,16); break;
				}
				read++;
			}

			cur = cur->next;
		}
	}

	fclose(fin);
	return read;
}


/**********************************************************************
	FUNCTION: ReadCfgPrint
	CONTENTS: This will print the cfg values to the file.
	PARAMETERS:
		filename -- The file to print to.  If this is null, output goes
			to the screen.
	RETURNS:
		nothing
	NOTES:
**********************************************************************/
void ReadCfgPrint(char *filename)
{
	FILE *fout;
	stCfgEntry *cur;

	if (filename == NULL) {
		fout = stdout;
	} else {
		if ((fout = fopen(filename,"w")) == NULL)
			return;
	}

	cur = cfgFirst;
	while (cur != NULL) {
		fprintf(fout, "%30.30s : ",cur->name);
		switch (cur->type) {
			case RC_TYPE_INT :
				fprintf (fout,"%12d (%12d)", *(int *) cur->var, cur->deflt.i);
				break;
			case RC_TYPE_LONG	:
				fprintf (fout,"%12ld (%12ld)", *(long *) cur->var, cur->deflt.l);
				break;
			case RC_TYPE_FLOAT :
				fprintf (fout,"%12.5g (%12.5g)", *(float *) cur->var, cur->deflt.f);
				break;
			case RC_TYPE_DOUBLE :
				fprintf (fout,"%12.5g (%12.5g)", *(double *) cur->var, cur->deflt.d);
				break;
			case RC_TYPE_HEX :
				fprintf (fout,"%12x (%12x)", *(int *) cur->var, cur->deflt.h);
				break;
		}
		fprintf(fout, "\n");
		cur = cur->next;
	}

	if (fout != stdout)
		fclose(fout);
}


/**********************************************************************
	FUNCTION: ReadCfgClear
	CONTENTS: This is called at the end of the cfg usage to free the
		memory used by the cfg stuff.
	PARAMETERS:
		none
	RETURNS:
		nothing
	NOTES:
**********************************************************************/
void ReadCfgClear(void)
{
	stCfgEntry *ne, *temp;

	ne = cfgFirst;

	while (ne != NULL) {
		temp = ne;
		ne = ne->next;
		free(temp->name);
		free(temp);
	}

	cfgFirst = cfgLast = NULL;
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/**********************************************************************
	FUNCTION: CreateCfgEntry
	CONTENTS: This will create a cfg entry with the given name and
		type and add it to the list.
	PARAMETERS:
		name -- The name for the entry.
		type -- The type of the entry.
		var -- The variable to access.
	RETURNS:
		stCfgEntry * -- The structure that was created.
	NOTES:
**********************************************************************/
stCfgEntry *CreateCfgEntry(char *name, int type, void *var)
{
	stCfgEntry *ne;

		/* allocate the structure */
	if ((ne = (stCfgEntry *) malloc(sizeof(stCfgEntry))) == NULL)
		return NULL;

		/* allocte the name entry */
	if ((ne->name = (char *) malloc(strlen(name)+1)) == NULL) {
		free(ne);
		return NULL;
	}

		/* fill the structure */
	strcpy(ne->name,name);
	ne->type = type;
	ne->var = var;
	ne->next = NULL;

			/* add it to the list */
	AddCfgEntry(ne);

	return ne;
}


/**********************************************************************
	FUNCTION: AddCfgEntry
	CONTENTS: This will add the entry to the list to read into.
	PARAMETERS:
		ne -- The new entry.
	RETURNS:
		nothing
	NOTES:
**********************************************************************/
void AddCfgEntry(stCfgEntry *ne)
{
	if (cfgFirst == NULL) {
		cfgFirst = ne;
		cfgLast = ne;
	} else {
		cfgLast->next = ne;
		cfgLast = ne;
	}
}